/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <algorithm>
#include <iostream>
#include <list>
#include <map>
#include <unordered_map>
#include <string>
#include <utility>
#include <vector>

#include "errors.h"
#include "sm_defs.h"

/* 字段元数据 */
struct ColMeta {
    std::string tab_name; // 字段所属表名称
    std::string name; // 字段名称
    ColType type; // 字段类型
    int len; // 字段长度
    int offset; // 字段位于记录中的偏移量
    bool index; /** unused */

    friend std::ostream &operator<<(std::ostream &os, const ColMeta &col) {
        // ColMeta中有各个基本类型的变量，然后调用重载的这些变量的操作符<<（具体实现逻辑在defs.h）
        return os << col.tab_name << ' ' << col.name << ' ' << col.type << ' ' << col.len << ' ' << col.offset << ' '
               << col.index;
    }

    friend std::istream &operator>>(std::istream &is, ColMeta &col) {
        return is >> col.tab_name >> col.name >> col.type >> col.len >> col.offset >> col.index;
    }

    // 重载相等运算符
    bool operator==(const ColMeta &other) const {
        return tab_name == other.tab_name && name == other.name;
    }
};

/* 索引元数据 */
struct IndexMeta {
    std::string tab_name; // 索引所属表名称
    int col_tot_len = 0; // 索引字段长度总和
    int col_num = 0; // 索引字段数量
    std::vector<std::pair<int, ColMeta> > cols; // 索引包含的字段 在索引中的偏移量 -> 列元信息
    std::unordered_map<std::string, std::pair<int, ColMeta> > cols_map;

    IndexMeta() = default;

    IndexMeta(std::string &&tab_name_, int col_tot_len_, int col_num_,
              std::vector<ColMeta> &&cols_) : tab_name(std::move(tab_name_)), col_tot_len(col_tot_len_),
                                              col_num(col_num_) {
        int offset = 0;
        for (auto &col: cols_) {
            cols.emplace_back(offset, std::move(col));
            offset += cols.back().second.len;
        }
        for (auto it = cols.begin(); it != cols.end(); ++it) {
            cols_map.emplace(it->second.name, *it);
        }
    }

    friend std::ostream &operator<<(std::ostream &os, const IndexMeta &index) {
        os << index.tab_name << " " << index.col_tot_len << " " << index.col_num;
        for (auto &[_, col]: index.cols) {
            os << "\n" << col;
        }
        return os;
    }

    friend std::istream &operator>>(std::istream &is, IndexMeta &index) {
        is >> index.tab_name >> index.col_tot_len >> index.col_num;
        int offset = 0;
        int len = 0;
        for (int i = 0; i < index.col_num; ++i) {
            ColMeta col;
            is >> col;
            len = col.len;
            index.cols.emplace_back(offset, std::move(col));
            offset += len;
        }
        for (auto it = index.cols.begin(); it != index.cols.end(); ++it) {
            index.cols_map.emplace(it->second.name, *it);
        }
        return is;
    }

    // 重载相等运算符
    bool operator==(const IndexMeta &other) const {
        return tab_name == other.tab_name && col_tot_len == other.col_tot_len &&
               col_num == other.col_num && cols == other.cols;
    }
};

// 定义自定义哈希函数
namespace std {
    template<>
    struct hash<IndexMeta> {
        std::size_t operator()(const IndexMeta &index) const {
            std::size_t hash = std::hash<std::string>{}(index.tab_name);
            hash ^= std::hash<int>{}(index.col_tot_len) << 1;
            hash ^= std::hash<int>{}(index.col_num) << 2;
            for (const auto &[_, col]: index.cols) {
                hash ^= std::hash<std::string>{}(col.name) << 3;
            }
            return hash;
        }
    };
}

/* 表元数据 */
struct TabMeta {
    std::string name; // 表名称
    std::vector<ColMeta> cols; // 表包含的字段
    std::unordered_map<std::string, IndexMeta> indexes; // 表上建立的索引 索引名 -> 索引元信息

    TabMeta() {
    }

    TabMeta(const TabMeta &other) {
        name = other.name;
        for (auto &col: other.cols) {
            cols.emplace_back(col);
        }
    }

    /* 判断当前表中是否存在名为col_name的字段 */
    bool is_col(const std::string &col_name) const {
        auto pos = std::find_if(cols.begin(), cols.end(), [&](const ColMeta &col) { return col.name == col_name; });
        return pos != cols.end();
    }

    std::string get_index_name(const std::vector<std::string> &index_cols) {
        std::string ix_name(name);
        for (const auto &col: index_cols) {
            name.append("_").append(col);
        }
        return std::move(name.append(".idx"));
    }

    /* 判断当前表上是否建有指定索引，索引包含的字段为col_names */
    bool is_index(const std::vector<std::string> &col_names) {
        // TODO 用哈希表优化
        auto &&ix_name = get_index_name(col_names);
        return indexes.count(ix_name) != 0;
        // for(auto& index: indexes) {
        //     if(index.col_num == col_names.size()) {
        //         size_t i = 0;
        //         for(; i < index.col_num; ++i) {
        //             if(index.cols[i].name.compare(col_names[i]) != 0)
        //                 break;
        //         }
        //         if(i == index.col_num) return true;
        //     }
        // }
    }

    /* 根据字段名称集合获取索引元数据 */
    IndexMeta &get_index_meta(const std::vector<std::string> &col_names) {
        // TODO 优化
        auto &&it = indexes.find(get_index_name(col_names));
        if (it != indexes.end()) {
            return it->second;
        }
        // for(auto index = indexes.begin(); index != indexes.end(); ++index) {
        //     if((*index).col_num != col_names.size()) continue;
        //     auto& index_cols = (*index).cols;
        //     size_t i = 0;
        //     for(; i < col_names.size(); ++i) {
        //         if(index_cols[i].name.compare(col_names[i]) != 0)
        //             break;
        //     }
        //     if(i == col_names.size()) return index;
        // }
        throw IndexNotFoundError(name, col_names);
    }

    /* 根据字段名称获取字段元数据 */
    std::vector<ColMeta>::iterator get_col(const std::string &col_name) {
        auto pos = std::find_if(cols.begin(), cols.end(), [&](const ColMeta &col) { return col.name == col_name; });
        if (pos == cols.end()) {
            throw ColumnNotFoundError(col_name);
        }
        return pos;
    }

    friend std::ostream &operator<<(std::ostream &os, const TabMeta &tab) {
        os << tab.name << '\n' << tab.cols.size() << '\n';
        for (auto &col: tab.cols) {
            os << col << '\n'; // col是ColMeta类型，然后调用重载的ColMeta的操作符<<
        }
        os << tab.indexes.size() << "\n";
        for (auto &[index_name, index]: tab.indexes) {
            os << index_name << '\n';
            os << index << "\n";
        }
        return os;
    }

    friend std::istream &operator>>(std::istream &is, TabMeta &tab) {
        size_t n;
        is >> tab.name >> n;
        for (size_t i = 0; i < n; i++) {
            ColMeta col;
            is >> col;
            tab.cols.emplace_back(col);
        }
        is >> n;
        std::string index_name;
        IndexMeta index;
        for (size_t i = 0; i < n; ++i) {
            is >> index_name;
            is >> index;
            tab.indexes.emplace(std::move(index_name), std::move(index));
            // TODO 注意这里要清空
            index.cols.clear();
            index.cols_map.clear();
        }
        return is;
    }
};

// 注意重载了操作符 << 和 >>，这需要更底层同样重载TabMeta、ColMeta的操作符 << 和 >>
/* 数据库元数据 */
class DbMeta {
    friend class SmManager;

private:
    std::string name_; // 数据库名称
    std::map<std::string, TabMeta> tabs_; // 数据库中包含的表

public:
    // DbMeta(std::string name) : name_(name) {}

    /* 判断数据库中是否存在指定名称的表 */
    bool is_table(const std::string &tab_name) const { return tabs_.find(tab_name) != tabs_.end(); }

    void SetTabMeta(const std::string &tab_name, const TabMeta &meta) {
        tabs_[tab_name] = meta;
    }

    /* 获取指定名称表的元数据 */
    TabMeta &get_table(const std::string &tab_name) {
        auto pos = tabs_.find(tab_name);
        if (pos == tabs_.end()) {
            throw TableNotFoundError(tab_name);
        }

        return pos->second;
    }

    // 重载操作符 <<
    friend std::ostream &operator<<(std::ostream &os, const DbMeta &db_meta) {
        os << db_meta.name_ << '\n' << db_meta.tabs_.size() << '\n';
        for (auto &entry: db_meta.tabs_) {
            os << entry.second << '\n';
        }
        return os;
    }

    friend std::istream &operator>>(std::istream &is, DbMeta &db_meta) {
        size_t n;
        is >> db_meta.name_ >> n;
        for (size_t i = 0; i < n; i++) {
            TabMeta tab;
            is >> tab;
            db_meta.tabs_[tab.name] = tab;
        }
        return is;
    }
};
